/**
 * \file
 * \copyright (C) 2016-2018 Alexey L. Voskov
 *
 * Laboratory of Chemical Thermodynamics, Department of Chemistry,
 *
 * Lomonosov Moscow State University, Moscow, Russia, 119991
 *
 * This file is the part of the CpFit program
 *
 * This program is free software; you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation; either version 2 of the License, or
 * (at your option) any later version.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 * You should have received a copy of the GNU General Public License
 * along with this program.  If not, see http://www.gnu.org/licenses/);
 */
#include <stdio.h>
#include <stdlib.h>
#include "wx/wx.h"
#include "wx/grid.h"
#include "wx/bookctrl.h"
#include "wx/textfile.h"
#include "wx/tokenzr.h"
#include "wx/aboutdlg.h"
#include "wx/filefn.h"
#include "datafit.h"
#include "plotcanvas.h"
#include "customgrids.h"
#include "cp.h"
#include "graph.h"
#include "misc.h"


/**
 * \brief Controls IDs
 */
enum {
	ID_BTN_CALC = 100,
	ID_CMB_PARNUM,
	ID_BTN_RESETPAR,
	ID_BTN_ADDPOINT,
	ID_BTN_DELPOINT,
	ID_BTN_SETREL,
	ID_BTN_SETABS,

	ID_GR_PARAM,
	ID_GR_DATA,
	// Fileinfo text fields (file metadata)
	ID_TXT_FILEINFO0,
	ID_TXT_FILEINFO1,
	ID_TXT_FILEINFO2,
	// Menu commands: general
	ID_MNU_NEW,
	ID_MNU_OPEN,
	ID_MNU_SAVEAS,
	ID_MNU_LOADPARAM,
	ID_MNU_EXIT,
	ID_MNU_ABOUT,
	ID_MNU_CUSTOMTERM,
	// Menu commands: calculation
	ID_MNU_CALC,
	ID_MNU_AUTOCALC,
	ID_MNU_CALC_ADDTERMS,
	ID_MNU_CALC_EXTERMSONLY,
	ID_MNU_CALC_NOOPT,
	// Menu commands: data operation
	ID_MNU_DATA_SORT
};



////////////////////////////////////////
wxBEGIN_EVENT_TABLE(MainWindow, wxFrame)
	EVT_MENU(ID_MNU_NEW, MainWindow::OnNew)
	EVT_MENU(ID_MNU_OPEN, MainWindow::OnOpenFile)
	EVT_MENU(ID_MNU_SAVEAS, MainWindow::OnSaveAsFile)
	EVT_MENU(ID_MNU_LOADPARAM, MainWindow::OnLoadParameters)
	EVT_MENU(ID_MNU_EXIT, MainWindow::OnExit)
	EVT_CLOSE(MainWindow::OnClose)
	EVT_MENU(ID_MNU_ABOUT, MainWindow::OnAbout)
	EVT_MENU(ID_MNU_CUSTOMTERM, MainWindow::OnAddCustomTerm)

	EVT_MENU(ID_MNU_CALC, MainWindow::OnEvalPar)
	EVT_MENU(ID_MNU_AUTOCALC, MainWindow::OnAutoEvalPar)
	EVT_MENU(ID_MNU_CALC_ADDTERMS, MainWindow::OnAddTermsEvalPar)
	EVT_MENU(ID_MNU_CALC_EXTERMSONLY, MainWindow::OnEvalExTermsOnly)
	EVT_MENU(ID_MNU_CALC_NOOPT, MainWindow::OnNoOptEvalPar)
	
	EVT_MENU(ID_MNU_DATA_SORT, MainWindow::OnDataSort)
	
	EVT_BUTTON(ID_BTN_RESETPAR, MainWindow::OnResetPar)
	EVT_BUTTON(ID_BTN_CALC, MainWindow::OnEvalPar)
	
	EVT_COMBOBOX(ID_CMB_PARNUM, MainWindow::OnCmbParNum)
	EVT_GRID_CELL_CHANGING(MainWindow::OnParamGridCellChanging)
	
	EVT_SIZE(MainWindow::OnSize)
	EVT_TEXT(ID_TXT_FILEINFO0, MainWindow::OnTextEvent)
	EVT_TEXT(ID_TXT_FILEINFO1, MainWindow::OnTextEvent)
	EVT_TEXT(ID_TXT_FILEINFO2, MainWindow::OnTextEvent)
wxEND_EVENT_TABLE()


/**
 * \brief  Creates all elements in the main window form
 */
MainWindow::MainWindow() : wxFrame(NULL, wxID_ANY, wxString("CpFit"))
{
	strInitApprox = wxString::FromUTF8("Initial approximation");
	strInputData = wxString::FromUTF8("Input data");
	strNumOfTerms = wxString::FromUTF8("Number of terms");
	strResetParams = wxString::FromUTF8("Reset parameters");

	strNotNumeric = wxString::FromUTF8("Not numeric parameter");

	wxGridCellBoolEditor::UseStringValues("1", "0");

	wxString parNumChoices[] = {"1", "2", "3", "4", "5", "6", "7", "8", "9", "10", "11", "12"};
	wxPanel *panel = new wxPanel(this, wxID_ANY);
	isFileChanged = false;
	SetIcon(wxICON(sample));
	//---------- Static box with file metadata ----------
	wxSize mdatLblSize(190, 25);
	wxSize mdatTxtSize(530, 25);
	wxPoint mdatLblPos(20, 30);
	wxPoint mdatTxtPos(210, 30);
	boxInfo = new wxStaticBox(panel, wxID_ANY, "File metadata", wxPoint(10, 10), wxSize(740, 150));
	lblFileInfo[0] = new wxStaticText(boxInfo, wxID_ANY, "File description 1", mdatLblPos, mdatLblSize); // Field 1 (lbl + txt)
	txtFileInfo[0] = new wxTextCtrl(boxInfo, ID_TXT_FILEINFO0, "Data set description", mdatTxtPos, mdatTxtSize);
	mdatLblPos.y += 40; mdatTxtPos.y += 40;
	lblFileInfo[1] = new wxStaticText(boxInfo, wxID_ANY, "File description 2", mdatLblPos, mdatLblSize); // Field 2 (lbl + txt)
	txtFileInfo[1] = new wxTextCtrl(boxInfo, ID_TXT_FILEINFO1, "T, K", mdatTxtPos, mdatTxtSize);
	mdatLblPos.y += 40; mdatTxtPos.y += 40;
	lblFileInfo[2] = new wxStaticText(boxInfo, wxID_ANY, "File description 3", mdatLblPos, mdatLblSize); // Field 3 (lbl + txt)
	txtFileInfo[2] = new wxTextCtrl(boxInfo, ID_TXT_FILEINFO2, "Cp, J/mol/K", mdatTxtPos, mdatTxtSize);
	mdatLblPos.y += 40; mdatTxtPos.y += 40;
	//---------- Static box with initial parameters ----------
	box = new wxStaticBox(panel, wxID_ANY, strInitApprox, wxPoint(10, 10+110+10), wxSize(300, 300));
	// a) grid with parameters
	grPar = new wxGrid(box, ID_GR_PARAM, wxPoint(20, 30), wxSize(230,200));
	grPar->CreateGrid(3, 2);
	grPar->SetColLabelValue(0, wxT("\u03B1")); // alpha
	grPar->SetColLabelValue(1, wxT("\u03B8 / K")); // theta
	grPar->SetRowLabelSize(40);
	grPar->SetColSize(0, 80);
	for (int i = 0; i < 3; i++) {
		grPar->SetCellValue(i, 0, "1e-6");
		grPar->SetCellValue(i, 1, "1");
	}
	// b) Selector of parameters number
	lblParNum = new wxStaticText(box, wxID_ANY, strNumOfTerms,
		wxPoint(10, 230), wxSize(150, 25));
	cmbParNum = new wxComboBox(box, ID_CMB_PARNUM, "3",
		wxPoint(170, 230), wxSize(50, 25), 12, parNumChoices, wxCB_READONLY);
	// c) Reset parameters button
	btnCalc = new wxButton(box, ID_BTN_CALC, "Evaluate parameters", wxPoint(10,260), wxSize(150,30));
	btnReset = new wxButton(box, ID_BTN_RESETPAR, strResetParams, wxPoint(170,260), wxSize(150,30));
	//---------- Static box with initial data ----------
	boxData = new wxStaticBox(panel, wxID_ANY, strInputData, wxPoint(10+300+10, 10+110+10), wxSize(430, 300));
	auto cbfFileChanged = [&isFileChanged] () { isFileChanged = true; }; // Callback for DataFileEditor
	ctlData = new DataFileEditor(boxData, cbfFileChanged, wxPoint(20, 30), wxSize(380, 230));
	// Window menu
	wxMenu *menuFile = new wxMenu();
	menuFile->Append(ID_MNU_NEW, "&New\tALT-N", "New data set");
	menuFile->AppendSeparator();
	menuFile->Append(ID_MNU_OPEN, "&Open\tALT-O", "Open data file");
	menuFile->Append(ID_MNU_SAVEAS, "Save as", "Save current data as...");
	menuFile->AppendSeparator();
	menuFile->Append(ID_MNU_LOADPARAM, "Load parameters", "Load alpha and theta parameters from .csv file");
	menuFile->Append(ID_MNU_CUSTOMTERM, "Load custom extra term", "Load custom extra Cp term (lua script)");
	menuFile->AppendSeparator();
	menuFile->Append(ID_MNU_EXIT, "E&xit\tALT-X", "Quit this program");
	wxMenu *menuEval = new wxMenu();
	menuEval->Append(ID_MNU_CALC, "Parameters values", "Parameters values");
	menuEval->AppendSeparator();
	menuEval->Append(ID_MNU_AUTOCALC, "Parameters set (don't use init.approx.)", "Parameters number and values");
	menuEval->Append(ID_MNU_CALC_ADDTERMS, "Parameters set (use init.approx.)", "Parameters numbers and values (use init.approx.)");
	menuEval->Append(ID_MNU_CALC_EXTERMSONLY, "Parameters set: extra terms only", "Parameters numbers and values: extra terms only");
	menuEval->AppendSeparator();
	menuEval->Append(ID_MNU_CALC_NOOPT, "Evaluate without optimization", "Evaluate without optimization");
	wxMenu *menuData = new wxMenu();
	menuData->Append(ID_MNU_DATA_SORT, "Sort data", "Sort data");
	wxMenu *menuHelp = new wxMenu();
	menuHelp->Append(ID_MNU_ABOUT, "About", "About the program");
	wxMenuBar *menuBar = new wxMenuBar();
	menuBar->Append(menuFile, "&File");
	menuBar->Append(menuEval, "&Evaluate");
	menuBar->Append(menuData, "&Data");
	menuBar->Append(menuHelp, "&Help");
	SetMenuBar(menuBar);
	
	
	SetSize(wxSize(900, 600));
	SetMinSize(wxSize(900, 600));
	RefreshTitle();
}

/**
 * \brief  Addition of custom terms (lua scripts)
 */
void MainWindow::OnAddCustomTerm(wxCommandEvent &event)
{
	// Create open file dialog (for .lua files)
	wxFileDialog openFileDialog(this, wxT("Open custom term script file"), "", "",
                       "lua script files (*.lua)|*.lua", wxFD_OPEN|wxFD_FILE_MUST_EXIST);
	// Run open file dialog and load the custom term from it
	if (openFileDialog.ShowModal() != wxID_CANCEL) {
		// Term loading process
		ExtraTermContainer::InitTermsTypes();
		wxString fileName = openFileDialog.GetPath();
		LuaTerm *term = new LuaTerm(std::string(fileName.c_str()));
		wxString title = wxT("Custom term loader");
		if (term->isValid) {		
			// Try to call the function
			double Cp = term->CpFuncEval((term->Tmin + term->Tmax) / 2);
			if (Cp != Cp) {
				wxMessageBox(term->GetErrorMessage(), title, wxICON_ERROR);
				return; // Cancel addition of term
			}
			// Add term
			if (ExtraTermContainer::AddExtraTermType(term)) {
				wxString msg = wxT("Term ") + term->GetType() + wxT(" successfully loaded");
				ctlData->RefreshExtraTermsTypes();
				wxMessageBox(msg, wxT("Custom term loader"), wxICON_INFORMATION);
			} else {
				wxMessageBox(wxT("Error during loading ") + term->GetErrorMessage(), title, wxICON_ERROR);
			}
		} else {
			wxMessageBox(term->GetErrorMessage(), title, wxICON_ERROR);
		}
	}
}


/**
 * \brief  Main window resizing
 */
void MainWindow::OnSize(wxSizeEvent &event)
{
	wxSize s = event.GetSize();
	// Set static boxes coordinates
	boxInfo->SetSize(s.x - 30, boxInfo->GetSize().y);
	int ymin = boxInfo->GetPosition().y + boxInfo->GetSize().y + 10;
	int totalWidth = s.x - 50;
	box->SetSize(totalWidth * 400 / 1000, s.y - ymin - 70);
	boxData->SetSize(totalWidth * 600 / 1000, s.y - ymin - 70);
	box->Move(box->GetPosition().x, ymin);
	boxData->Move(20 + box->GetPosition().x + box->GetSize().x, ymin);
	// Set coordinates of elements inside boxInfo (metadata)
	s = boxInfo->GetSize();
	for (int i = 0; i < 3; i++) {
		txtFileInfo[i]->SetSize(s.x - txtFileInfo[i]->GetPosition().x - 20, 25);
	}
	// Set coordinates of elements inside box (parameters)
	s = box->GetSize();
	// a) buttons
	btnCalc->Move(btnCalc->GetPosition().x, s.y - btnCalc->GetSize().y - 10);
	btnReset->Move(btnReset->GetPosition().x, s.y - btnReset->GetSize().y - 10);
	// b) label + combo box
	lblParNum->Move(lblParNum->GetPosition().x, btnCalc->GetPosition().y - lblParNum->GetSize().y - 10);
	cmbParNum->Move(cmbParNum->GetPosition().x, btnCalc->GetPosition().y - lblParNum->GetSize().y - 10);
	// c) grid (table)
	grPar->SetSize(s.x - 70, lblParNum->GetPosition().y - 40);
	// Set coordinates of elements inside boxData (experimental data)
	s = boxData->GetSize();
	ctlData->SetSize(s.x - 40, s.y - 40);
	event.Skip();	
	// Resize columns inside grPar and gr Data tables
	ResizeWxGridCols(grPar);
}

/**
 * \brief Parameters numbers (i.e. number of terms) selection (by means of combo box)
 */
void MainWindow::OnCmbParNum(wxCommandEvent &event)
{
  	ResizeParamGrid(cmbParNum->GetSelection() + 1);
	grPar->SetFocus();
}

/**
 * \brief  Controls validity of user data editing
 */
void MainWindow::OnParamGridCellChanging(wxGridEvent& event)
{
	if (event.GetId() == ID_GR_PARAM) {
		double t;
		if (!event.GetString().ToDouble(&t)) {
			wxMessageBox(strNotNumeric, "CpFit", wxICON_ERROR);
			event.Veto();
		}
	}
	event.Skip();
}

/**
 * \brief  Creation of new data set (reset everyting)
 */
void MainWindow::OnNew(wxCommandEvent &event)
{
	OnResetPar(event);
	ctlData->Reset();
	dataFile.Reset();
	txtFileInfo[0]->ChangeValue("Data set description");
	txtFileInfo[1]->ChangeValue("T, K");
	txtFileInfo[2]->ChangeValue("Cp, J/mol/K");	
	SetIsFileChanged(false);
}

/**
 * \brief Open the file and load data set from it
 */
void MainWindow::OnOpenFile(wxCommandEvent &event)
{    
	// Check if file has been changed (to prevent data loss)
	if (isFileChanged) {
		int ans = wxMessageBox("Current file is not saved. All changes will be lost. Continue?", "CpFit",
			wxYES_NO | wxNO_DEFAULT | wxICON_EXCLAMATION);
		if (ans == wxNO) {
			return; // The user cancelled opening the file: stop it!
		}
	}
	// Create open file dialog (for .dat files)
	wxFileDialog openFileDialog(this, wxT("Open dat file"), "", "",
                       "dat files (*.dat;*.txt)|*.dat;*.txt", wxFD_OPEN|wxFD_FILE_MUST_EXIST);
	// Run open file dialog and open the file                       
	if (openFileDialog.ShowModal() != wxID_CANCEL) {
		wxString fileName = openFileDialog.GetPath();
		if (dataFile.LoadFile(fileName)) {
			RefreshDataGrid();
			SetIsFileChanged(false);
		} else {
			wxMessageBox("Cannot load the file", "CpFit", wxICON_ERROR | wxOK);
		}
	}
}

/**
 * Save current data set to file
 */
void MainWindow::OnSaveAsFile(wxCommandEvent &event)
{
	// Synchronize window with data
	ctlData->SaveToDataFile(dataFile);
	std::vector <wxString> fileInfo;
	for (int i = 0; i < 3; i++) {
		fileInfo.push_back(txtFileInfo[i]->GetValue());
	}
	dataFile.SetFileInfo(fileInfo);
	// Get file name
	wxFileDialog openFileDialog(this, wxT("Save dat file"), "", "",
                       "dat files (*.dat;*.txt)|*.dat;*.txt", wxFD_SAVE);
	// Save file on disk
	if (openFileDialog.ShowModal() != wxID_CANCEL) {
		wxString fileName = openFileDialog.GetPath();
		// Checks if the file exists (if exists -- ask if we need 
		// to overwrite it)
		wxTextFile file;
		bool writeToFile = true;
		if (wxFileExists(fileName)) {
			wxString msg = wxString::Format("File %s already exists. Overwrite it?", fileName);
			if (wxMessageBox(msg, "CpFit", wxICON_EXCLAMATION | wxYES_NO | wxNO_DEFAULT) == wxNO) {
				writeToFile = false;
			}
		}
		// Save to the file (if allowed by the user)
		if (writeToFile) {
			dataFile.alpha_0 = ReadAlpha();
			dataFile.theta_0 = ReadTheta();
			if (dataFile.SaveFile(fileName)) {
				SetIsFileChanged(false);
			} else {
				wxMessageBox("Cannot save the file", "CpFit", wxICON_ERROR | wxOK);
			}
		}
	}
}

/**
 * \brief Loads model parameters from CSV file
 *
 * The first row contains alpha and theta strings,
 * the rest of rows are parameters. Dot is decimal point,
 * semicolon is separator.
 */
void MainWindow::OnLoadParameters(wxCommandEvent &event)
{
	wxFileDialog openFileDialog(this, wxT("Load parameters values"), "", "",
                       "csv files (*.csv)|*.csv", wxFD_OPEN);
	if (openFileDialog.ShowModal() != wxID_CANCEL) {
		// Open file
		wxString fileName = openFileDialog.GetPath();
		wxTextFile file;		
		if (!file.Open(fileName)) {
			wxString msg = wxString::Format("Cannot open file %s", fileName);
			wxMessageBox(msg, "CpFit", wxICON_ERROR);
			return;
		}
		// Read file header
		// a) Check header presence
		if (file.Eof()) {
			file.Close();
			wxMessageBox("File is empty!", "CpFit", wxICON_ERROR);
			return;
		}
		// b) find alpha and theta in the header (its indices)
		int alpha_ind = -1, theta_ind = -1, ind = 0;
		wxString str = file.GetFirstLine();
		wxStringTokenizer tokenizer(str, ";");
		while (tokenizer.HasMoreTokens()) {
			wxString token = tokenizer.GetNextToken();
			if (token == "alpha") {
				alpha_ind = ind;
			} else if (token == "theta") {
				theta_ind = ind;
			}
			ind++;
		}
		int numOfCols = ind;
		// c) check if alpha and theta fields exist
		if (alpha_ind == -1) {
			file.Close();
			wxMessageBox("Cannot find alpha column", "CpFit", wxICON_ERROR);
			return;
		}
		if (theta_ind == -1) {
			file.Close();
			wxMessageBox("Cannot find theta column", "CpFit", wxICON_ERROR);
			return;
		}
		// Load parameters values
		std::vector <double> alpha_new, theta_new;
		for (str = file.GetNextLine(); !file.Eof(); str = file.GetNextLine()) {
			// Find and load alpha and theta values from the row
			tokenizer = wxStringTokenizer(str, ";");
			int rowInd = 0;
			bool alphaLoaded = false, thetaLoaded = false;
			while (tokenizer.HasMoreTokens()) {
				// Get new token
				wxString token = tokenizer.GetNextToken();
				// Convert token to double
				double val;
				if (!token.ToDouble(&val)) {
					file.Close();
					wxString msg = wxString::Format("%s is not a number", token);
					wxMessageBox(msg, "CpFit", wxICON_ERROR);
					return;
				}
				// Load values to vectors
				if (rowInd == alpha_ind) {
					alpha_new.push_back(val);
					alphaLoaded = true;
				}
				if (rowInd == theta_ind) {
					theta_new.push_back(val);
					thetaLoaded = true;
				}
				rowInd++;
			}
			// Check if alpha and theta values have been loaded
			if (!alphaLoaded) {
				wxMessageBox("Cannot find alpha value!", "CpFit", wxICON_ERROR);
			}
			if (!thetaLoaded) {
				wxMessageBox("Cannot find theta value!", "CpFit", wxICON_ERROR);
			}
		}
		// Save it into CpFit forms
		WriteAlpha(alpha_new);
		WriteTheta(theta_new);
	}	
}

/**
 * \brief Show "About the program" dialog
 */
void MainWindow::OnAbout(wxCommandEvent &event)
{
	wxString GPLText("This program is free software: you can redistribute it and/or modify "
    "it under the terms of the GNU General Public License as published by "
    "the Free Software Foundation, either version 2 of the License, or "
    "(at your option) any later version.\n\n"
    "This program is distributed in the hope that it will be useful, "
    "but WITHOUT ANY WARRANTY; without even the implied warranty of "
    "MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the "
    "GNU General Public License for more details.\n\n"
    "You should have received a copy of the GNU General Public License "
    "along with this program.  If not, see <http://www.gnu.org/licenses/>.");

	wxAboutDialogInfo info;
	info.SetName(("CpFit"));
	info.SetVersion(_("0.8"));
	info.SetDescription(_(
		"This program describes heat capacities experimental data\n"
		"using Einstein-Plank Cp(T) empirical functions\n"
		"and the nonlinear least squares method\n"));
	info.SetCopyright(wxT("(C) 2016-2018 Laboratory of Chemical Thermodynamics\nLomonosow Moscow State University"));
	info.SetWebSite("http://td.chem.msu.ru");
	info.AddDeveloper("Alexey L. Voskov (alvoskov@gmail.com)");
	info.SetLicense(GPLText);
	wxAboutBox(info);
}


/**
 * \brief  Reset all parameters values to default (i.e. default initial approximation)
 */
void MainWindow::OnResetPar(wxCommandEvent &event)
{
	for (int i = 0; i < grPar->GetNumberRows(); i++) {
		grPar->SetCellValue(i, 0, "1e-6");
		grPar->SetCellValue(i, 1, "1");
	}
}

/**
 * \brief  Run parameters evaluation (by means of the least squares method)
 */
void MainWindow::OnEvalPar(wxCommandEvent &event)
{
	ctlData->SaveToDataFile(dataFile);
	if (dataFile.GetNumOfPoints() > 0) {
		GraphWindow *grwnd = new GraphWindow((MainWindow *) this, "CpFit results");	
		grwnd->AddData(dataFile);
		grwnd->AddParam0(ReadAlpha(), ReadTheta());
		grwnd->CurveFit();
		grwnd->Show( true );
	} else {
		wxMessageBox("No data are present", "CpFit", wxOK | wxICON_EXCLAMATION);
	}
}

/**
 * \brief Run parameters evaluation using autodefinition of the parameters number.
 * User-defined initial approximation will be ignored! It is based on the least
 * squares method.
 */
void MainWindow::OnAutoEvalPar(wxCommandEvent &event)
{
	ctlData->SaveToDataFile(dataFile);
	if (dataFile.GetNumOfPoints() > 0) {
		GraphWindow *grwnd = new GraphWindow((MainWindow *) this, "CpFit results");	
		grwnd->AddData(dataFile);
		grwnd->CurveFit(FIT_VALNTERMS_AUTO);
		grwnd->Show( true );
	} else {
		wxMessageBox("No data are present", "CpFit", wxOK | wxICON_EXCLAMATION);
	}
}

/**
 * \brief Run parameters evaluation using user-defined initial approximation but
 * adds extra terms while they are statistically significant (can be used when
 * MainWindow::OnAutoEvalPar stops at small number of terms, i.e. in local minimum)
 */
void MainWindow::OnAddTermsEvalPar(wxCommandEvent &event)
{
	ctlData->SaveToDataFile(dataFile);
	if (dataFile.GetNumOfPoints() > 0) {
		GraphWindow *grwnd = new GraphWindow((MainWindow *) this, "CpFit results");	
		grwnd->AddData(dataFile);
		grwnd->AddParam0(ReadAlpha(), ReadTheta());
		grwnd->CurveFit(FIT_VALNTERMS_INITAPPROX);
		grwnd->Show( true );
	} else {
		wxMessageBox("No data are present", "CpFit", wxOK | wxICON_EXCLAMATION);
	}
}

/**
 * \brief Run parameters evaluation only for extra terms (baseline parameters are not
 * changing). In this case "initial approximation" parameters are used unchanged.
 */
void MainWindow::OnEvalExTermsOnly(wxCommandEvent &event)
{
	ctlData->SaveToDataFile(dataFile);
	if (dataFile.GetNumOfPoints() > 0) {
		GraphWindow *grwnd = new GraphWindow((MainWindow *) this, "CpFit results");	
		grwnd->AddData(dataFile);
		grwnd->AddParam0(ReadAlpha(), ReadTheta()); // Will be used as baseline WITHOUT CHANGES!
		grwnd->CurveFit(FIT_EXTERMS_ONLY);
		grwnd->Show( true );
	} else {
		wxMessageBox("No data are present", "CpFit", wxOK | wxICON_EXCLAMATION);
	}
}

/**
 * \brief Shows graphs but doesn't optimize the coefficients. Can work either with
 * or without experimental data
 */
void MainWindow::OnNoOptEvalPar(wxCommandEvent &event)
{
	ctlData->SaveToDataFile(dataFile);
	GraphWindow *grwnd = new GraphWindow((MainWindow *) this, "CpFit results");
	grwnd->AddData(dataFile);
	grwnd->AddParam0(ReadAlpha(), ReadTheta());
	grwnd->CurveFit(FIT_NOOPT);
	grwnd->Show( true );
}

/**
 * \brief Sorts all experimental data by reference, by type and by T
 */
void MainWindow::OnDataSort(wxCommandEvent &event)
{
	dataFile.SortData();	
	RefreshDataGrid();
	SetIsFileChanged(true);
}


/**
 * \brief Mainly to control editing text in metadata as changing the file
 */
void MainWindow::OnTextEvent(wxCommandEvent& event)
{
	switch (event.GetId()) {
	case ID_TXT_FILEINFO0:
		SetIsFileChanged(true);
		break;
	case ID_TXT_FILEINFO1:
		SetIsFileChanged(true);
		break;
	case ID_TXT_FILEINFO2:
		SetIsFileChanged(true);
		break;

	}
	event.Skip();
}



/**
 * \brief Some operations before closing the window
 */
void MainWindow::OnExit(wxCommandEvent &event)
{
	Close(true);
}


/**
 * \brief Asks questions about not saved data before closing
 */
void MainWindow::OnClose(wxCloseEvent& event)
{
	if (event.CanVeto() && isFileChanged) {
		int ans = wxMessageBox("Current file is not saved. Exit without changes?", "CpFit",
			wxYES_NO | wxNO_DEFAULT | wxICON_EXCLAMATION);
		if (ans == wxNO) {
			event.Veto();
			return;
		}
	}
	event.Skip();
}

/**
 * \brief Load existing data to the grid
 */
void MainWindow::RefreshDataGrid()
{
	std::vector <wxString> strFileInfo = dataFile.GetFileInfo();
	if (strFileInfo.size() == 3) {
		for (int i = 0; i < 3; i++) {
			txtFileInfo[i]->ChangeValue(strFileInfo[i]);
		}
	}
	ctlData->LoadDataFile(dataFile);
	// Load initial approximations (if presented)
	if (dataFile.alpha_0.size() != 0) {
		WriteAlpha(dataFile.alpha_0);
		WriteTheta(dataFile.theta_0);
	}
}

/**
 * \brief Sets isFileChanged flag and then updates the window title
 * \param value New boolean value
 */
void MainWindow::SetIsFileChanged(bool value)
{
	isFileChanged = value;
	if (value == false) {
		RefreshTitle();
	}
}


/**
 * \brief  Change number of parameters (terms) inside of the grid
 * \param  Number of terms (table rows)
 */
void MainWindow::ResizeParamGrid(int numOfRows)
{
	int numOfRows_old = grPar->GetNumberRows();
	if (numOfRows > numOfRows_old) {
		grPar->InsertRows(numOfRows_old, numOfRows - numOfRows_old);
		for (int i = numOfRows_old; i < numOfRows; i++) {
			grPar->SetCellValue(i, 0, "1e-6");
			grPar->SetCellValue(i, 1, "1");
		}
	} else if (numOfRows < numOfRows_old) {
		grPar->DeleteRows(numOfRows, numOfRows_old - numOfRows);
	}
  	cmbParNum->SetSelection(numOfRows - 1);
}

/**
 * \brief  Refreshes window title using file name
 */
void MainWindow::RefreshTitle()
{
	wxString fileName = dataFile.GetFileName();
	if (fileName == "") {
		fileName = "<untitled>";
	}
	SetTitle(wxString::Format("CpFit: %s", fileName));
}



/**
 * \brief  Collect all alpha values from the grid into a vector and return them
 */
std::vector <double> MainWindow::ReadAlpha()
{
	int numOfRows = grPar->GetNumberRows();
	std::vector <double> alpha;
	for (int i = 0; i < numOfRows; i++) {
		double a;
		grPar->GetCellValue(i, 0).ToDouble(&a); // TODO:ERROR CHECK
		alpha.push_back(a);
	}
	return alpha;
}

/**
 * \brief  Collect all theta values from the grid into a vector and return them
 */
std::vector <double> MainWindow::ReadTheta()
{
	int numOfRows = grPar->GetNumberRows();
	std::vector <double> theta;
	for (int i = 0; i < numOfRows; i++) {
		double t;
		grPar->GetCellValue(i, 1).ToDouble(&t); // TODO:ERROR CHECK
		theta.push_back(t);
	}
	return theta;
}

/**
 * \brief  Write alpha values to the table
 *
 * The table will be resized to make sizes of alpha and table grid consistent.
 */
bool MainWindow::WriteAlpha(std::vector <double> alpha)
{
	int numOfRows = alpha.size();
	ResizeParamGrid(numOfRows);
	for (int i = 0; i < numOfRows; i++) {
		grPar->SetCellValue(i, 0, wxString::Format("%.14g", alpha[i]));
	}
	return true;
}

/**
 * \brief  Write theta values to the table
 *
 * The table will be resized to make sizes of alpha and table grid consistent.
 */
bool MainWindow::WriteTheta(std::vector <double> theta)
{
	int numOfRows = theta.size();
	ResizeParamGrid(numOfRows);
	for (int i = 0; i < numOfRows; i++) {
		grPar->SetCellValue(i, 1, wxString::Format("%.14g", theta[i]));
	}
	return true;
}


/**
 * \brief  Write extra terms parameters values to the data file (in memory)
 *
 * The function controls parameters types
 */
bool MainWindow::WriteExtraTermsParams(std::vector <ExtraTermContainer> &extraTerms_new)
{
	if (dataFile.extraTerms.size() != extraTerms_new.size()) {
		return false;
	}
	for (int i = 0; i < extraTerms_new.size(); i++) {
		if (!extraTerms_new[i].CompareTo(dataFile.extraTerms[i])) {
			return false;
		}
	}
	dataFile.extraTerms = extraTerms_new;
	ctlData->LoadDataFile(dataFile);
	return true;
}


/**
 * \brief  Adds data point to the data file
 */
void MainWindow::AddDataPoint(Point point)
{
	dataFile.data.push_back(point);
	RefreshDataGrid();
	SetIsFileChanged(true);
}

/**
 * \brief  Adds data points to the data file
 */
void MainWindow::AddDataPoints(std::vector <Point> points)
{
	for (int i = 0; i < points.size(); i++) {
		dataFile.data.push_back(points[i]);
	}
	RefreshDataGrid();
	SetIsFileChanged(true);
}



/**
 * \brief CpFit main class
 */
class MyApp: public wxApp
{
public:
	virtual bool OnInit();
};

wxIMPLEMENT_APP(MyApp);

/**
 * \brief CpFit entry point
 */
bool MyApp::OnInit()
{
	


	wxInitAllImageHandlers();
	MainWindow *frame = new MainWindow();
	frame->Show( true );
	return true;
}
