/**
 * \file 
 * \brief Contains classes that implement extra terms for non-linear regression
 *
 * \author Alexey L. Voskov (alvoskov@gmail.com)
 * \copyright (C) 2016-2018 Alexey L. Voskov
 *
 * Laboratory of Chemical Thermodynamics, Department of Chemistry,
 *
 * Lomonosov Moscow State University, Moscow, Russia, 119991
 *
 * This file is the part of the CpFit program
 *
 * This program is free software; you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation; either version 2 of the License, or
 * (at your option) any later version.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 * You should have received a copy of the GNU General Public License
 * along with this program.  If not, see http://www.gnu.org/licenses/);
 */

#include "datafit.h"
#include "dfunc3.h"
#include <string>
#include <vector>
#include <map>
#include <stdio.h>
#include <stdlib.h>
#include <memory.h>
#include <math.h>

//////////////////////////////////////////////////////////////////////////////////////////
ExtraTermDefault::ExtraTermDefault()
{
	Tmin = 280; Tmax = 320;
	// Default parameters values
	lsqparam.push_back(0.0); lsqparam_lb.push_back(0); lsqparam_ub.push_back(1e5);
	lsqparam.push_back(0.0); lsqparam_lb.push_back(0); lsqparam_ub.push_back(1e5);
	lsqparam.push_back(0.0); lsqparam_lb.push_back(-1.0); lsqparam_ub.push_back(1.0);
	fixparam.push_back(300.0);
	// Parameters names
	lsqparam_names.push_back("b1");
	lsqparam_names.push_back("b2");
	lsqparam_names.push_back("b3");
	fixparam_names.push_back("Ttr");
	// Term type ID	
	type = "DEFAULT";
	briefDescr = "Default";
	fullDescr = "Default term for lambda-transitions";
}

ExtraTermDefault::~ExtraTermDefault()
{
}


double ExtraTermDefault::CpFunc(double T, std::vector <double> &p, std::vector <double> &c, double Tmin, double Tmax)
{
	double dT = T - c[0];
	return p[0] * exp(p[1]*(p[2]*dT - fabs(dT))) * CONST_R;
}

double ExtraTermDefault::HFuncDefTerm(double T, double *p, double *c, double Tmin, double Tmax)
{
	if (T < Tmin) {
		return 0.0;
	} else {
		double T1 = (Tmin - 10 > 1e-3) ? (Tmin - 10) : 1e-3; 
		double T2 = (T > Tmax + 10) ? (Tmax + 10) : T;
		double Ttr = c[0], A = p[0], B, H = 0;
				
		if (T2 <= Ttr) {
			B = p[1]*(p[2] + 1);
			H = A * ( exp(B*(T2 - Ttr)) - exp(B*(T1 - Ttr)) ) / B;
		} else {
			// Left part (below Ttr)
			B = p[1]*(p[2] + 1);
			H = A * (1 - exp(B*(T1 - Ttr)) ) / B;
			// Right part (above Ttr)					
			B = p[1]*(p[2] - 1);
			H += A * ( exp(B*(T2 - Ttr)) - 1 ) / B;
		}
		return H * CONST_R;
	}
}

double ExtraTermDefault::HFunc(double T, std::vector <double> &p, std::vector <double> &c, double Tmin, double Tmax)
{
	return HFuncDefTerm(T, &p[0], &c[0], Tmin, Tmax);
}

double ExtraTermDefault::SFuncDefTerm(double T, double *p, double *c, double Tmin, double Tmax)
{
	if (T < Tmin) {
		return 0.0;		
	} else {
		double T1 = (Tmin - 10 > 1e-3) ? (Tmin - 10) : 1e-3; 
		double T2 = (T > Tmax + 10) ? (Tmax + 10) : T;
		double Ttr = c[0], A = p[0], B;
		// Left part (below Ttr)
		B = p[1]*(p[2] + 1);
		double S = A * (eminusx_ei(B*std::min(T2, Ttr)) - eminusx_ei(B*T1)*exp(B*(T1-Ttr)) );
		if (T2 > Ttr) { // Right part (above Ttr)					
			B = p[1]*(p[2] - 1);
			S += A * (eminusx_ei(B*T2)*exp(B*(T2 - Ttr)) - eminusx_ei(B*Ttr) );
		}
		return S * CONST_R;
	}
}

double ExtraTermDefault::SFunc(double T, std::vector <double> &p, std::vector <double> &c, double Tmin, double Tmax)
{
	return SFuncDefTerm(T, &p[0], &c[0], Tmin, Tmax);
}

ExtraTerm * ExtraTermDefault::Clone()
{
	ExtraTermDefault *clone = new ExtraTermDefault();
	clone->CopyData(this);
	return clone;	
}


/////////////////////////////////////////////////////////////////////////////////////////
ExtraTermTwoPeaks::ExtraTermTwoPeaks()
{
	Tmin = 280; Tmax = 330;
	// Default parameters values
	lsqparam.push_back(0.0); lsqparam_lb.push_back(0); lsqparam_ub.push_back(1e5);
	lsqparam.push_back(0.0); lsqparam_lb.push_back(0); lsqparam_ub.push_back(1e5);
	lsqparam.push_back(0.0); lsqparam_lb.push_back(-1.0); lsqparam_ub.push_back(1.0);
	lsqparam.push_back(0.0); lsqparam_lb.push_back(0); lsqparam_ub.push_back(1e5);
	lsqparam.push_back(0.0); lsqparam_lb.push_back(0); lsqparam_ub.push_back(1e5);
	lsqparam.push_back(0.0); lsqparam_lb.push_back(-1.0); lsqparam_ub.push_back(1.0);
	fixparam.push_back(300.0);
	fixparam.push_back(310.0);
	// Parameters names
	lsqparam_names.push_back("d1(1)");
	lsqparam_names.push_back("d2(1)");
	lsqparam_names.push_back("d3(1)");
	lsqparam_names.push_back("d1(2)");
	lsqparam_names.push_back("d2(2)");
	lsqparam_names.push_back("d3(2)");
	fixparam_names.push_back("Ttr1");
	fixparam_names.push_back("Ttr2");
	// Term type ID	
	type = "TWOPEAKS";
	briefDescr = "Two peaks";
	fullDescr = "Two peaks term for two lambda transitions";
}

ExtraTermTwoPeaks::~ExtraTermTwoPeaks()
{
}


double ExtraTermTwoPeaks::CpFunc(double T, std::vector <double> &p, std::vector <double> &c, double Tmin, double Tmax)
{
	double dT1 = T - c[0];
	double dT2 = T - c[1];
	return p[0] * exp(p[1]*(p[2]*dT1 - fabs(dT1))) * CONST_R + 
	       p[3] * exp(p[4]*(p[5]*dT2 - fabs(dT2))) * CONST_R;
}

double ExtraTermTwoPeaks::HFunc(double T, std::vector <double> &p, std::vector <double> &c, double Tmin, double Tmax)
{
	return ExtraTermDefault::HFuncDefTerm(T, &p[0], &c[0], Tmin, Tmax) + 
	       ExtraTermDefault::HFuncDefTerm(T, &p[3], &c[1], Tmin, Tmax);
}

double ExtraTermTwoPeaks::SFunc(double T, std::vector <double> &p, std::vector <double> &c, double Tmin, double Tmax)
{
	return ExtraTermDefault::SFuncDefTerm(T, &p[0], &c[0], Tmin, Tmax) + 
		ExtraTermDefault::SFuncDefTerm(T, &p[3], &c[1], Tmin, Tmax);
}

ExtraTerm * ExtraTermTwoPeaks::Clone()
{
	ExtraTermTwoPeaks *clone = new ExtraTermTwoPeaks();
	clone->CopyData(this);
	return clone;	
}


////////////////////////////////////////////////////////////////////////////////////////
ExtraTermLeftExp::ExtraTermLeftExp()
{
	Tmin = 280; Tmax = 300;
	// Default parameters values
	lsqparam.push_back(1.0); lsqparam_lb.push_back(0); lsqparam_ub.push_back(1e5);
	lsqparam.push_back(1.0); lsqparam_lb.push_back(0); lsqparam_ub.push_back(1e5);
	// Parameters names
	lsqparam_names.push_back("a(1)");
	lsqparam_names.push_back("p(1)");
	// Term type ID	
	type = "LEFTEXP";
	briefDescr = "Left exp.";
	fullDescr = "Exponent-based function for left part of lambda peak";
}

ExtraTermLeftExp::~ExtraTermLeftExp()
{
}

double ExtraTermLeftExp::CpFuncLExpTerm(double T, double a, double p, double Tmax)
{
	return (T <= Tmax) ? CONST_R * T * a * exp(p * (T - Tmax)) : 0;
}

double ExtraTermLeftExp::HFuncLExpTerm(double T, double a, double p, double Tmax)
{
	if (T > Tmax) {
		T = Tmax;
	}
	return CONST_R * a / (p * p) * ( exp(p * (T - Tmax)) * (T * p - 1)  + exp(- p * Tmax) );	
}

double ExtraTermLeftExp::SFuncLExpTerm(double T, double a, double p, double Tmax)
{
	if (T > Tmax) {
		T = Tmax;
	}
	return CONST_R * a / p * (exp(p * (T - Tmax)) - exp(-p * Tmax));
}

double ExtraTermLeftExp::CpFunc(double T, std::vector <double> &p, std::vector <double> &c, double Tmin, double Tmax)
{
	return ExtraTermLeftExp::CpFuncLExpTerm(T, p[0], p[1], Tmax);
}

double ExtraTermLeftExp::HFunc(double T, std::vector <double> &p, std::vector <double> &c, double Tmin, double Tmax)
{
	return ExtraTermLeftExp::HFuncLExpTerm(T, p[0], p[1], Tmax);
}

double ExtraTermLeftExp::SFunc(double T, std::vector <double> &p, std::vector <double> &c, double Tmin, double Tmax)
{
	return ExtraTermLeftExp::SFuncLExpTerm(T, p[0], p[1], Tmax);
}

ExtraTerm * ExtraTermLeftExp::Clone()
{
	ExtraTermLeftExp *clone = new ExtraTermLeftExp();
	clone->CopyData(this);
	return clone;	
}


/////////////////////////////////////////////////////////////////////////////////////
ExtraTermRightExp::ExtraTermRightExp()
{
	Tmin = 280; Tmax = 300;
	// Default parameters values
	lsqparam.push_back(1.0); lsqparam_lb.push_back(0); lsqparam_ub.push_back(1e5);
	lsqparam.push_back(1.0); lsqparam_lb.push_back(0); lsqparam_ub.push_back(1e5);
	// Parameters names
	lsqparam_names.push_back("b(1)");
	lsqparam_names.push_back("q(1)");
	// Term type ID	
	type = "RIGHTEXP";
	briefDescr = "Right exp.";
	fullDescr = "Exponent-based function for right part of lambda peak";
}

ExtraTermRightExp::~ExtraTermRightExp()
{
}

double ExtraTermRightExp::CpFuncRExpTerm(double T, double b, double q, double Tmin)
{
	return (T >= Tmin) ? CONST_R * T * b * exp(-q*(T - Tmin)) : 0;
}

double ExtraTermRightExp::HFuncRExpTerm(double T, double b, double q, double Tmin)
{
	return (T >= Tmin) ? CONST_R * b / (q * q) * (-exp(-q * (T - Tmin)) * (T * q + 1) + (Tmin * q + 1)) : 0;
}

double ExtraTermRightExp::SFuncRExpTerm(double T, double b, double q, double Tmin)
{
	return (T >= Tmin) ? CONST_R * (b / q) * (-exp(-q * (T - Tmin)) + 1)  : 0;
}


double ExtraTermRightExp::CpFunc(double T, std::vector <double> &p, std::vector <double> &c, double Tmin, double Tmax)
{
	return ExtraTermRightExp::CpFuncRExpTerm(T, p[0], p[1], Tmin);
}

double ExtraTermRightExp::HFunc(double T, std::vector <double> &p, std::vector <double> &c, double Tmin, double Tmax)
{
	return ExtraTermRightExp::HFuncRExpTerm(T, p[0], p[1], Tmin);
}

double ExtraTermRightExp::SFunc(double T, std::vector <double> &p, std::vector <double> &c, double Tmin, double Tmax)
{
	return ExtraTermRightExp::SFuncRExpTerm(T, p[0], p[1], Tmin);
}

ExtraTerm * ExtraTermRightExp::Clone()
{
	ExtraTermRightExp *clone = new ExtraTermRightExp();
	clone->CopyData(this);
	return clone;	
}


/////////////////////////////////////////////////////////////////////////////
ExtraTermLeftExp2::ExtraTermLeftExp2()
{
	Tmin = 280; Tmax = 300;
	// Default parameters values
	lsqparam.push_back(1.0); lsqparam_lb.push_back(0); lsqparam_ub.push_back(1e5);
	lsqparam.push_back(1.0); lsqparam_lb.push_back(0); lsqparam_ub.push_back(1e5);
	lsqparam.push_back(1.0); lsqparam_lb.push_back(0); lsqparam_ub.push_back(1e5);
	lsqparam.push_back(1.0); lsqparam_lb.push_back(0); lsqparam_ub.push_back(1e5);
	// Parameters names
	lsqparam_names.push_back("a(1)");
	lsqparam_names.push_back("p(1)");
	lsqparam_names.push_back("a(2)");
	lsqparam_names.push_back("p(2)");
	// Term type ID	
	type = "LEFTEXP2";
	briefDescr = "Left exp.2";
	fullDescr = "Exponent-based function for left part of lambda peak";
}

ExtraTermLeftExp2::~ExtraTermLeftExp2()
{
}

double ExtraTermLeftExp2::CpFunc(double T, std::vector <double> &p, std::vector <double> &c, double Tmin, double Tmax)
{
	return ExtraTermLeftExp::CpFuncLExpTerm(T, p[0], p[1], Tmax) +
		ExtraTermLeftExp::CpFuncLExpTerm(T, p[2], p[3], Tmax);
}

double ExtraTermLeftExp2::HFunc(double T, std::vector <double> &p, std::vector <double> &c, double Tmin, double Tmax)
{
	return ExtraTermLeftExp::HFuncLExpTerm(T, p[0], p[1], Tmax) + 
		ExtraTermLeftExp::HFuncLExpTerm(T, p[2], p[3], Tmax);
}

double ExtraTermLeftExp2::SFunc(double T, std::vector <double> &p, std::vector <double> &c, double Tmin, double Tmax)
{
	return ExtraTermLeftExp::SFuncLExpTerm(T, p[0], p[1], Tmax) + 
		ExtraTermLeftExp::SFuncLExpTerm(T, p[2], p[3], Tmax);
}

ExtraTerm * ExtraTermLeftExp2::Clone()
{
	ExtraTermLeftExp2 *clone = new ExtraTermLeftExp2();
	clone->CopyData(this);
	return clone;	
}


//////////////////////////////////////////////////////////////////////////////////////
ExtraTermRightExp2::ExtraTermRightExp2()
{
	Tmin = 280; Tmax = 300;
	// Default parameters values
	lsqparam.push_back(1.0); lsqparam_lb.push_back(0); lsqparam_ub.push_back(1e5);
	lsqparam.push_back(1.0); lsqparam_lb.push_back(0); lsqparam_ub.push_back(1e5);
	lsqparam.push_back(1.0); lsqparam_lb.push_back(0); lsqparam_ub.push_back(1e5);
	lsqparam.push_back(1.0); lsqparam_lb.push_back(0); lsqparam_ub.push_back(1e5);
	// Parameters names
	lsqparam_names.push_back("b(1)");
	lsqparam_names.push_back("q(1)");
	lsqparam_names.push_back("b(2)");
	lsqparam_names.push_back("q(2)");
	// Term type ID	
	type = "RIGHTEXP2";
	briefDescr = "Right exp.2";
	fullDescr = "Exponent-based function for right part of lambda peak";
}

ExtraTermRightExp2::~ExtraTermRightExp2()
{
}

double ExtraTermRightExp2::CpFunc(double T, std::vector <double> &p, std::vector <double> &c, double Tmin, double Tmax)
{
	return ExtraTermRightExp::CpFuncRExpTerm(T, p[0], p[1], Tmin) +
		ExtraTermRightExp::CpFuncRExpTerm(T, p[2], p[3], Tmin);
}

double ExtraTermRightExp2::HFunc(double T, std::vector <double> &p, std::vector <double> &c, double Tmin, double Tmax)
{
	return ExtraTermRightExp::HFuncRExpTerm(T, p[0], p[1], Tmin) + 
		ExtraTermRightExp::HFuncRExpTerm(T, p[2], p[3], Tmin);
}

double ExtraTermRightExp2::SFunc(double T, std::vector <double> &p, std::vector <double> &c, double Tmin, double Tmax)
{
	return ExtraTermRightExp::SFuncRExpTerm(T, p[0], p[1], Tmin) + 
		ExtraTermRightExp::SFuncRExpTerm(T, p[2], p[3], Tmin);
}

ExtraTerm * ExtraTermRightExp2::Clone()
{
	ExtraTermRightExp2 *clone = new ExtraTermRightExp2();
	clone->CopyData(this);
	return clone;	
}


////////////////////////////////////////////////////////////////////////////////////
ExtraTermSkewed::ExtraTermSkewed()
{
	Tmin = 280; Tmax = 320;
	// Default parameters values
	lsqparam.push_back(10.0);  lsqparam_lb.push_back(-1e5); lsqparam_ub.push_back(1e5);
	lsqparam.push_back(0.0);   lsqparam_lb.push_back(-1e5); lsqparam_ub.push_back(1e5);
	lsqparam.push_back(300.0); lsqparam_lb.push_back(0);    lsqparam_ub.push_back(1e5);
	lsqparam.push_back(100.0); lsqparam_lb.push_back(0);    lsqparam_ub.push_back(1e5);
	// Parameters names
	lsqparam_names.push_back("b(1)");
	lsqparam_names.push_back("b(2)");
	lsqparam_names.push_back("b(3)");
	lsqparam_names.push_back("b(4)");
	// Term type ID	
	type = "SKEWED";
	briefDescr = "Skewed";
	fullDescr = "Gauss-like peaks (with possible asymmetry)";
}

ExtraTermSkewed::~ExtraTermSkewed()
{
}

ExtraTerm * ExtraTermSkewed::Clone()
{
	ExtraTermSkewed *clone = new ExtraTermSkewed();
	clone->CopyData(this);
	return clone;	
}

double ExtraTermSkewed::CpFunc(double T, std::vector <double> &p, std::vector <double> &c, double Tmin, double Tmax)
{
	double x = (T - p[2]) / p[3];
	return CONST_R * p[0] * exp(-x*x) / (1 + exp(-p[1]*x));
}

////////////////////////////////////////////////////////////////////////////////////
ExtraTermEmpty::ExtraTermEmpty()
{
	Tmin = 280; Tmax = 320;
	// Default parameters values
	lsqparam.push_back(0); lsqparam_lb.push_back(-1); lsqparam_ub.push_back(1);
	// Parameters names
	lsqparam_names.push_back("a");
	// Term type ID	
	type = "EMPTY";
	briefDescr = "Empty";
	fullDescr = "Always 0";
}

ExtraTermEmpty::~ExtraTermEmpty()
{
}

ExtraTerm * ExtraTermEmpty::Clone()
{
	ExtraTermEmpty *clone = new ExtraTermEmpty();
	clone->CopyData(this);
	return clone;
}

double ExtraTermEmpty::CpFunc(double T, std::vector <double> &p, std::vector <double> &c, double Tmin, double Tmax)
{
	return 0.0;
}

double ExtraTermEmpty::HFunc(double T, std::vector <double> &p, std::vector <double> &c, double Tmin, double Tmax)
{
	return 0.0;
}

double ExtraTermEmpty::SFunc(double T, std::vector <double> &p, std::vector <double> &c, double Tmin, double Tmax)
{
	return 0.0;
}

////////////////////////////////////////////////////////////////////////////////////////
/**
 * @brief Creates a copy of LuaTerm class. No reading of the original file is required for it.
 */
ExtraTerm * LuaTerm::Clone()
{
	// Convert the function to the HEX dump (and save it in the buffer)
	InitLuaWriter();
	lua_getfield(L, -1, "cpfunc");
	lua_dump(L, LuaTerm::LuaWriter, (void *) this, 0);
	lua_pop(L, 1);	
	// Upload the function into new state
	LuaTerm *clone = new LuaTerm("..");
	clone->isValid = true;
	clone->type = this->type;
	clone->isValid = clone->InitLuaState();
	lua_createtable(clone->L, 0, 1);
	lua_pushstring(clone->L, "cpfunc");
	luaL_loadbuffer(clone->L, luaDumpBuf, luaDumpLen, "cpfunc");
	lua_settable(clone->L, -3);
	// Free buffers
	FreeLuaWriter();
	// Copy the rest of the data and return clone
	clone->CopyData(this);
	return clone;
}

/**
 * @brief lua_Writer function for internal use (to transfer Lua functions form one lua_State to other)
 */
int LuaTerm::LuaWriter(lua_State *L, const void *p, size_t sz, void *ud)
{
	LuaTerm *term = (LuaTerm *) ud;
	if (term->luaDumpBuf == NULL) {
		term->luaDumpBuf = (char *) malloc(sz);
		term->luaDumpLen = sz;
		memcpy(term->luaDumpBuf, p, sz);
	} else {
		term->luaDumpBuf = (char *) realloc(term->luaDumpBuf, term->luaDumpLen + sz);
		memcpy(term->luaDumpBuf + term->luaDumpLen, p, sz);
		term->luaDumpLen += sz;
	}
	return 0;
}

/**
 * @brief Initialize variables for buffer for LuaWriter (set everything to 0)
 */
void LuaTerm::InitLuaWriter()
{
	luaDumpBuf = NULL;
	luaDumpLen = 0;
}

/**
 * @brief Free buffer for LuaWriter
 */
void LuaTerm::FreeLuaWriter()
{
	if (luaDumpBuf) {
		free(luaDumpBuf);
		luaDumpLen = 0;
		luaDumpBuf = NULL;
	}
}

/**
 * @brief Initializes Lua state and loads custom term from .lua file
 * @param fileName Full name of Lua script with function.
 */
LuaTerm::LuaTerm(std::string fileName)
{
	if (fileName != "..") {
		type = "CUSTOM";
		this->fileName = fileName;
		if (InitLuaState()) {
			isValid = InitUserFuncData() ? true : false;
		} else {
			isValid = false;
		}
	} else {
		type = "..";
		isValid = false;
	}
}

LuaTerm::~LuaTerm()
{
	lua_close(L);
}

void LuaTerm::BufferToErrorMessage()
{
	errorMessage = std::string(errmsg);
}


bool LuaTerm::IsFieldTable(const char *fieldName)
{
	lua_getfield(L, -1, fieldName);
	if (lua_isnil(L, -1) || !lua_istable(L, -1)) {
		snprintf(errmsg, MSG_BUFSIZE, "%s field is absent or doesn't contain table", fieldName);
		BufferToErrorMessage();
		lua_pop(L, 1);
		return false;
	} else {
		lua_pop(L, 1);
		return true;
	}
}

int LuaTerm::GetFieldLen(const char *fieldName)
{
	lua_getfield(L, -1, fieldName);
	lua_len(L, -1);
	int n = (int) lua_tonumber(L, -1);
	lua_pop(L, 2);
	return n;
}

std::vector <double> LuaTerm::LuaTableToVector()
{
	std::vector <double> vec;
	// Get number of elements in the table
	lua_len(L, -1);
	int len = (int) lua_tonumber(L, -1);
	lua_pop(L, 1);
	// Convert table to std::vector
	for (int i = 1; i <= len; i++) {
		lua_geti(L, -1, i);
		vec.push_back(lua_tonumber(L, -1));
		lua_pop(L, 1);
	}
	// Return results	
	return vec;
}

bool LuaTerm::VectorToLuaTable(std::vector <double> &vec)
{
	int len = vec.size();
	lua_createtable(L, len, 0);
	for (int i = 0; i < len; i++) {
		lua_pushnumber(L, vec[i]);
		lua_seti(L, -2, i + 1);
	}
}

std::vector <std::string> LuaTerm::LuaTableToStrings()
{
	std::vector <std::string> vec;
	// Get number of elements in the table
	lua_len(L, -1);
	int len = (int) lua_tonumber(L, -1);
	lua_pop(L, 1);
	// Convert table to std::vector
	for (int i = 1; i <= len; i++) {
		size_t slen;
		lua_geti(L, -1, i);
		if (lua_isstring(L, -1)) {
			const char *sptr = lua_tolstring(L, -1, &slen);
			vec.push_back(std::string(sptr, slen));
		} else {
			vec.push_back(std::string(""));
		}
		lua_pop(L, 1);
	}
	// Return results	
	return vec;	
}


/**
 * @brief Call heat capacity function `cpfunc` from preloaded Lua script
 */
double LuaTerm::CpFunc(double T, std::vector <double> &p, std::vector <double> &c, double Tmin, double Tmax)
{	
	lua_getfield(L, -1, "cpfunc");
	if (lua_isnil(L, -1) || !lua_isfunction(L, -1)) {
		snprintf(errmsg, MSG_BUFSIZE, "Cannot find cpfunc");
		BufferToErrorMessage();
		return nan("");
	}
	// cpfunc(T, lsqparam, fixparam, Tmin, Tmax)
	lua_pushnumber(L, T);
	VectorToLuaTable(p);
	VectorToLuaTable(c);
	lua_pushnumber(L, Tmin);
	lua_pushnumber(L, Tmax);	
	if (lua_pcall(L, 5, 1, 0) != 0) {
		snprintf(errmsg, MSG_BUFSIZE, "Error in cpfunc %s", lua_tostring(L, -1));
		BufferToErrorMessage();
		lua_pop(L, 1);
		return nan("");
	}
	double Cp = lua_tonumber(L, -1);
	lua_pop(L, 1);
	return Cp;

}


/**
 * @brief Initialize Lua state with all required constants and C functions
 */
bool LuaTerm::InitLuaState()
{
	// New Lua state
	L = luaL_newstate();
	luaL_openlibs(L);
	// Load pre-defined constants
	lua_pushnumber(L, CONST_R);
	lua_setglobal(L, "R");
	lua_pushnumber(L, CONST_CAL);
	lua_setglobal(L, "CAL");
	lua_pushcfunction(L, LuaTerm::CpDebyeFunc);
	lua_setglobal(L, "CpDebye");
	// Prohibit creation of global variables
	// a) get table of globals
	lua_getglobal(L, "_G");
	// b) create metatable for it
	lua_newtable(L);
	lua_pushcfunction(L, LuaTerm::NoNewGlobalsFunc);
	lua_setfield(L, -2, "__newindex");
	// c) set metatable for it
	lua_setmetatable(L, -2);
	lua_pop(L, -1);
	return true;
}


/**
 * @brief Try to load user-defined function and model parameters from .lua script
 */
bool LuaTerm::InitUserFuncData()
{
	// Load user-defined function
	if (luaL_dofile(L, fileName.c_str()) != 0) {
		snprintf(errmsg, MSG_BUFSIZE, "Cannot load user-defined function (%s)", lua_tostring(L, -1));
		BufferToErrorMessage();
		return false;
	}
	
	if (lua_gettop(L) != 1 || lua_type(L, 1) != LUA_TTABLE) {
		snprintf(errmsg, MSG_BUFSIZE, "User-defined .lua file must return a table");
		BufferToErrorMessage();
		return false;
	}
	// Check data types
	lua_getfield(L, -1, "cpfunc");
	if (lua_isnil(L, -1) || !lua_isfunction(L, -1)) {
		snprintf(errmsg, MSG_BUFSIZE, "cpfunc field is absent or doesn't contain function");
		BufferToErrorMessage();
		lua_pop(L, 1);
		return false;
	}
	lua_pop(L, 1);
	// Try to get description
	lua_getfield(L, -1, "description");
	if (lua_isnil(L, -1) || !lua_isstring(L, -1)) {
		snprintf(errmsg, MSG_BUFSIZE, "description field is absent or is not string");
		BufferToErrorMessage();
		lua_pop(L, 1);
		return false;
	} else {
		size_t slen;
		const char *sptr = lua_tolstring(L, -1, &slen);
		briefDescr = std::string(sptr, slen);
		fullDescr = briefDescr;
		lua_pop(L, 1);
	}
	// Try to get formula
	lua_getfield(L, -1, "formula");
	if (lua_isnil(L, -1) || !lua_isstring(L, -1)) {
		snprintf(errmsg, MSG_BUFSIZE, "formula field is absent or is not string");
		BufferToErrorMessage();
		lua_pop(L, 1);
		return false;
	} else {
		size_t slen;
		const char *sptr = lua_tolstring(L, -1, &slen);
		formula = std::string(sptr, slen);
		lua_pop(L, 1);
	}

	// Check if there is a table 
	if (lua_gettop(L) == 0 || !lua_istable(L, -1)) {
		snprintf(errmsg, MSG_BUFSIZE, "The script doesn't return a table");
		BufferToErrorMessage();
		return false;
	}
	// Check fields types
	if (!IsFieldTable("lsqparam")) return false;
	if (!IsFieldTable("lsqparam_lb")) return false;
	if (!IsFieldTable("lsqparam_ub")) return false;
	if (!IsFieldTable("lsqparam_names")) return false;
	if (!IsFieldTable("fixparam")) return false;
	if (!IsFieldTable("fixparam_names")) return false;
	// Get Data type
	lua_getfield(L, -1, "type");
	if (lua_isnil(L, -1) || !lua_isstring(L, -1)) {
		snprintf(errmsg, MSG_BUFSIZE, "type field is absent or is not string");
		BufferToErrorMessage();
		lua_pop(L, 1);
		return false;
	} else {
		size_t slen;
		const char *sptr = lua_tolstring(L, -1, &slen);
		type = std::string(sptr, slen);
		lua_pop(L, 1);
	}
	// Get Tmin and Tmax values
	lua_getfield(L, -1, "Tmin");
	if (lua_isnil(L, -1) || !lua_isnumber(L, -1)) {
		snprintf(errmsg, MSG_BUFSIZE, "Tmin field is absent or is not numeric");
		BufferToErrorMessage();
		lua_pop(L, 1);
		return false;
	} else {
		Tmin = lua_tonumber(L, -1);
		lua_pop(L, 1);
	}
	lua_getfield(L, -1, "Tmax");
	if (lua_isnil(L, -1) || !lua_isnumber(L, -1)) {
		snprintf(errmsg, MSG_BUFSIZE, "Tmax field is absent or is not numeric");
		BufferToErrorMessage();
		lua_pop(L, 1);
		return false;
	} else {
		Tmax = lua_tonumber(L, -1);
		lua_pop(L, 1);
	}
	// Check tables sizes
	int nlsqparam, nfixparam, nparams;
	nlsqparam = GetFieldLen("lsqparam");
	nfixparam = GetFieldLen("fixparam");
	if (nlsqparam != GetFieldLen("lsqparam_lb")) {
		snprintf(errmsg, MSG_BUFSIZE, "Invalid size of lsqparam_lb table");
		BufferToErrorMessage();
		return false;		
	}
	if (nlsqparam != GetFieldLen("lsqparam_ub")) {
		snprintf(errmsg, MSG_BUFSIZE, "Invalid size of lsqparam_ub table");
		BufferToErrorMessage();
		return false;		
	}
	if (nlsqparam != GetFieldLen("lsqparam_names")) {
		snprintf(errmsg, MSG_BUFSIZE, "Invalid size of lsqparam_names table");
		BufferToErrorMessage();
		return false;		
	}
	if (nfixparam != GetFieldLen("fixparam_names")) {
		snprintf(errmsg, MSG_BUFSIZE, "Invalid size of fixparam_names table");
		BufferToErrorMessage();
		return false;		
	}
	// Load tables values
	lua_getfield(L, -1, "lsqparam");
	lsqparam = LuaTableToVector();
	lua_pop(L, 1);	

	lua_getfield(L, -1, "lsqparam_lb");
	lsqparam_lb = LuaTableToVector();
	lua_pop(L, 1);

	lua_getfield(L, -1, "lsqparam_ub");
	lsqparam_ub = LuaTableToVector();
	lua_pop(L, 1);	

	lua_getfield(L, -1, "lsqparam_names");
	lsqparam_names = LuaTableToStrings();
	lua_pop(L, 1);
	
	lua_getfield(L, -1, "fixparam");
	fixparam = LuaTableToVector();
	lua_pop(L, 1);	

	lua_getfield(L, -1, "fixparam_names");
	fixparam_names = LuaTableToStrings();
	lua_pop(L, 1);

	return true;
}

/**
 * @brief Lua function for Debye Cp model, i.e. Cp = CpDebyeFunc(T, TD)
 */
int LuaTerm::CpDebyeFunc(lua_State *L)
{
	// Check and get input arguments
	int nargin = lua_gettop(L);
	if (nargin != 2) {
           lua_pushstring(L, "Incorrect number of arguments");
           lua_error(L);		
	}
	if (!lua_isnumber(L, -2)) {
           lua_pushstring(L, "T must be numeric");
           lua_error(L);		
	}
	if (!lua_isnumber(L, -1)) {
           lua_pushstring(L, "TD must be numeric");
           lua_error(L);		
	}
	double T = lua_tonumber(L, -2);
	double TD = lua_tonumber(L, -1);
	// Calculate function
	double Cp_D = 3.0*CONST_R*c_debye_func(TD / T);
	// Return it
	lua_pushnumber(L, Cp_D);
	return 1;	
}

/**
 * @brief Implementation of __newindex metamethod for metatable for _G
 * that prohibits creation of global variables
 */
int LuaTerm::NoNewGlobalsFunc(lua_State *L)
{
	char buf[MSG_BUFSIZE];
	lua_getglobal(L, "debug");
	lua_getfield(L, -1, "traceback");
	lua_pcall(L, 0, 1, 0);
	snprintf(buf, MSG_BUFSIZE, "no creation of global variables is allowed\n%s", lua_tostring(L, -1));
	lua_pop(L, 3);
	lua_pushstring(L, buf);
	lua_error(L);
	return 0;
}
