/**
 * \file 
 * \brief Contains classes that implement extra terms for non-linear regression
 *
 * \author Alexey L. Voskov (alvoskov@gmail.com)
 * \copyright (C) 2016-2018 Alexey L. Voskov
 *
 * Laboratory of Chemical Thermodynamics, Department of Chemistry,
 *
 * Lomonosov Moscow State University, Moscow, Russia, 119991
 *
 * This file is the part of the CpFit program
 *
 * This program is free software; you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation; either version 2 of the License, or
 * (at your option) any later version.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 * You should have received a copy of the GNU General Public License
 * along with this program.  If not, see http://www.gnu.org/licenses/);
 */


#ifndef __DATAFIT_ETERMS_H
#define __DATAFIT_ETERMS_H
#include "lua/lua.hpp"

/**
 * @brief Extra term for lambda-like transition with possible asymmetry 
 *
 * \f[
 * \frac{C_p^\textrm{ex}}{R} = b_1\exp\left[b_2\left(b_3\Delta T - \left|
 * \Delta T \right|\right)\right];
 * ~
 * \Delta T = T - T_\textrm{tr};
 * ~
 * T_\textrm{tr} = c_1
 * \f]
 */
class ExtraTermDefault : public ExtraTerm
{
public:
	ExtraTermDefault();
	~ExtraTermDefault();
	virtual ExtraTerm * Clone();
	
	static double HFuncDefTerm(double T, double *p, double *c, double Tmin, double Tmax);
	static double SFuncDefTerm(double T, double *p, double *c, double Tmin, double Tmax);	
protected:
	virtual double CpFunc(double T, std::vector <double> &p, std::vector <double> &c, double Tmin, double Tmax);
	virtual double HFunc(double T, std::vector <double> &p, std::vector <double> &c, double Tmin, double Tmax);
	virtual double SFunc(double T, std::vector <double> &p, std::vector <double> &c, double Tmin, double Tmax);
	
};

/**
 * @brief Extra term for two partially overlapping lambda-like transitions with possible asymmetry
 *
 * \f[
 * \frac{C_p^\textrm{ex}}{R} = \sum_{i=1}^2 d_1^{(i)}\exp\left[d_2^{(i)}\left(d_3^{(i)}(\Delta T)_i - \left|
 * (\Delta T)_i \right|\right)\right];
 * ~
 * (\Delta T)_i = T - T_{\textrm{tr},i};
 * ~
 * T_{\textrm{tr},i} = c_i
 * \f]
 * \f[
 * d_i^{(j)} = b_{i + 3(j-1)}
 * \f]
 */
class ExtraTermTwoPeaks : public ExtraTerm
{
public:
	ExtraTermTwoPeaks();
	~ExtraTermTwoPeaks();
	virtual ExtraTerm * Clone();
protected:
	virtual double CpFunc(double T, std::vector <double> &p, std::vector <double> &c, double Tmin, double Tmax);
	virtual double HFunc(double T, std::vector <double> &p, std::vector <double> &c, double Tmin, double Tmax);
	virtual double SFunc(double T, std::vector <double> &p, std::vector <double> &c, double Tmin, double Tmax);	
};

/**
 * @brief Extra term for approximation of the left part of lambda-like transitions
 *
 * \f[
 * \frac{C_p^\textrm{ex}}{R} = b_1 \exp\left[b_2\left(T - T_\textrm{max}\right)\right]
 * \f]
 */
class ExtraTermLeftExp : public ExtraTerm
{
public:
	ExtraTermLeftExp();
	~ExtraTermLeftExp();
	virtual ExtraTerm * Clone();

	static double CpFuncLExpTerm(double T, double a, double p, double Tmax);
	static double HFuncLExpTerm(double T, double a, double p, double Tmax);
	static double SFuncLExpTerm(double T, double a, double p, double Tmax);	
protected:
	virtual double CpFunc(double T, std::vector <double> &p, std::vector <double> &c, double Tmin, double Tmax);
	virtual double HFunc(double T, std::vector <double> &p, std::vector <double> &c, double Tmin, double Tmax);
	virtual double SFunc(double T, std::vector <double> &p, std::vector <double> &c, double Tmin, double Tmax);	
};


/**
 * @brief Extra term for approximation of the right part of lambda-like transition
 *
 * \f[
 * \frac{C_p^\textrm{ex}}{R} = b_1 \exp\left[-b_2\left(T - T_\textrm{min}\right)\right]
 * \f]
 */
class ExtraTermRightExp : public ExtraTerm
{
public:
	ExtraTermRightExp();
	~ExtraTermRightExp();
	virtual ExtraTerm * Clone();

	static double CpFuncRExpTerm(double T, double b, double q, double Tmin);
	static double HFuncRExpTerm(double T, double b, double q, double Tmin);
	static double SFuncRExpTerm(double T, double b, double q, double Tmin);	
protected:
	virtual double CpFunc(double T, std::vector <double> &p, std::vector <double> &c, double Tmin, double Tmax);
	virtual double HFunc(double T, std::vector <double> &p, std::vector <double> &c, double Tmin, double Tmax);
	virtual double SFunc(double T, std::vector <double> &p, std::vector <double> &c, double Tmin, double Tmax);	
};


/**
 * @brief Extra term for approximation of the left part of lambda-like transition (more complex variant)
 *
 * \f[
 * \frac{C_p^\textrm{ex}}{R} =
 * b_1 \exp\left[b_2\left(T - T_\textrm{max}\right)\right] + 
 * b_3 \exp\left[b_4\left(T - T_\textrm{max}\right)\right]		
 * \f]
 */
class ExtraTermLeftExp2 : public ExtraTerm
{
public:
	ExtraTermLeftExp2();
	~ExtraTermLeftExp2();
	virtual ExtraTerm * Clone();

protected:
	virtual double CpFunc(double T, std::vector <double> &p, std::vector <double> &c, double Tmin, double Tmax);
	virtual double HFunc(double T, std::vector <double> &p, std::vector <double> &c, double Tmin, double Tmax);
	virtual double SFunc(double T, std::vector <double> &p, std::vector <double> &c, double Tmin, double Tmax);	
};


/**
 * @brief Extra term for approximation of the right part of lambda-like transition (more complex variant)
 *
 * \f[
 * \frac{C_p^\textrm{ex}}{R} =
 * b_1 \exp\left[-b_2\left(T - T_\textrm{min}\right)\right] + 
 * b_3 \exp\left[-b_4\left(T - T_\textrm{min}\right)\right]
 * \f]
 */
class ExtraTermRightExp2 : public ExtraTerm
{
public:
	ExtraTermRightExp2();
	~ExtraTermRightExp2();	
	virtual ExtraTerm * Clone();

protected:
	virtual double CpFunc(double T, std::vector <double> &p, std::vector <double> &c, double Tmin, double Tmax);
	virtual double HFunc(double T, std::vector <double> &p, std::vector <double> &c, double Tmin, double Tmax);
	virtual double SFunc(double T, std::vector <double> &p, std::vector <double> &c, double Tmin, double Tmax);	
};


/**
 * @brief Extra term for approximation of peaks resembling Gaussian curve (with possible asymmetry)
 *
 * \f[
 * \frac{C_p^\textrm{ex}}{R} = 
 * \frac{b_1 e^{-x^2}}{1 + e^{-b_2x}};~x = \frac{T - b_3}{b_4}
 * \f]
 */
class ExtraTermSkewed : public ExtraTerm
{
public:
	ExtraTermSkewed();
	~ExtraTermSkewed();	
	virtual ExtraTerm * Clone();

protected:
	virtual double CpFunc(double T, std::vector <double> &p, std::vector <double> &c, double Tmin, double Tmax);
};

/**
 * Extra terms that always returns 0 (zero). Useful for simple exclusion of anomalies.
 */
class ExtraTermEmpty : public ExtraTerm
{
public:
	ExtraTermEmpty();
	~ExtraTermEmpty();	
	virtual ExtraTerm * Clone();

protected:
	virtual double CpFunc(double T, std::vector <double> &p, std::vector <double> &c, double Tmin, double Tmax);
	virtual double HFunc(double T, std::vector <double> &p, std::vector <double> &c, double Tmin, double Tmax);
	virtual double SFunc(double T, std::vector <double> &p, std::vector <double> &c, double Tmin, double Tmax);	
};


/**
 * @brief  Allows to load custom extra Cp terms written in Lua
 *
 * Lua script must return table with the next fields:
 * - `cpfunc` -- function for calculation of heat capacity. The next function is expected:
 *   `cp = cpfunc(T, lsqparam, fixparam, Tmin, Tmax)`
 * - `lsqparam`, `lsqparam_lb`, `lsqparam_ub` -- parameters for LSQ (initial approximation,
 *   lower boundaries, upper boundaries). Each value is a table of doubles.
 * - `lsqparam_names` -- names of parameters in `lsqparam`. Table of strings.
 * - `fixparam`, `fixparam_names` -- fixed parameters and their names (format is similar
 *   to lsqparam and `lsqparam_names`)
 * - `Tmin`, `Tmax` -- minimal and maximal temperature, K
 * - `type` -- string with parameter type (will be used by CpFit; used to distinguish between
 *   different Lua scripts)
 * - `description` -- string with brief textual description (will be used in combobox)
 * - `formula` -- string with formula (UTF-8 letters and `^` and `_` symbols for uppercases
 *   and lowercases are supported
 *
 */
class LuaTerm : public ExtraTerm
{
public:
	LuaTerm(std::string fileName);
	~LuaTerm();
	ExtraTerm *Clone();

	bool isValid;	
private:
	bool VectorToLuaTable(std::vector <double> &vec);
	double CpFunc(double T, std::vector <double> &p, std::vector <double> &c, double Tmin, double Tmax);
	static int CpDebyeFunc(lua_State *L);
	static int NoNewGlobalsFunc(lua_State *L);
	void InitLuaWriter();
	static int LuaWriter(lua_State *L, const void *p, size_t sz, void *ud);
	void FreeLuaWriter();
	void BufferToErrorMessage();	

	bool IsFieldTable(const char *fieldName);
	int GetFieldLen(const char *fieldName);
	std::vector <std::string> LuaTableToStrings();
	std::vector <double> LuaTableToVector();
	bool InitLuaState();
	bool InitUserFuncData();


	std::string fileName;
	static const int MSG_BUFSIZE = 1024;
	char errmsg[MSG_BUFSIZE];

	char *luaDumpBuf;
	size_t luaDumpLen;
	
	lua_State *L;
};

#endif