/**
 * \brief Genetic algorithm for heat capacities fitting
 */
class LSQGeneticFit {
public:
	LSQGeneticFit(LSQContext &lsq, int nparams);
	std::vector <double> DataFit();
	std::vector <double> DataFit(std::vector <double> param0);

	bool verbose;
private:
	LSQContext lsqCont; /*!< Experimental data */
	int nparams; /*!< Number of model parameters */
	int popSize; /*!< Size of the population */
	int eliteSize; /*!< Size of the elite of the population */
	int maxIter; /*!< Maximal number of iterations */

	double std_alpha;
	double std_theta;
};


/*===== LSQGeneticFit class implementation =====*/
double randn()
{
	double phi, r, z;
	phi = (double) rand() / RAND_MAX;
	if (phi == 0) phi = 1e-10;
	r = (double) rand() / RAND_MAX;
	if (r == 0) r = 1e-10;
	z = cos(2*M_PI*phi)*sqrt(-2*log(r));
	return z;
}


class ParamVector {
public:
	std::vector <double> p;
	double rss;
	bool operator< (const ParamVector &right) const
	{
		return this->rss < right.rss;
	}

	void SetParam(std::vector <double> &newp)
	{
		p = newp;
	}

	void Mutate(double std_alpha, double std_theta)
	{
		for (int i = 0; i < p.size(); i += 2) {
			p[i] += randn() * std_alpha;
			if (p[i] < 1e-8) p[i] = fabs(p[i]) + 1e-8;
			p[i+1] += randn() * std_theta;
			if (p[i+1] < 1e-8) p[i+1] = fabs(p[i+1]) + 1e-8;
		}
	}

	void Mutate(double *std_alpha, double *std_theta)
	{
		for (int i = 0; i < p.size(); i += 2) {
			p[i] += randn() * std_alpha[i / 2];
			if (p[i] < 1e-8) p[i] = fabs(p[i]) + 1e-8;
			p[i+1] += randn() * std_theta[i / 2];
			if (p[i+1] < 1e-8) p[i+1] = fabs(p[i+1]) + 1e-8;
		}
	}


	void Cross(const ParamVector &p1, const ParamVector &p2)
	{
		for (int i = 0; i < p.size(); i++) {
			double w = rand() / RAND_MAX;
			p[i] = (1 - w)*p1.p[i] + w * p2.p[i];
		}
	}

	void CalcRSS(LSQContext &lsq)
	{
		TDPEFunc tdf(&p[0], p.size());
		std::vector <double> res = lsq.GetResVector(tdf);
		rss = 0.0;
		for (int i = 0; i < res.size(); i++) {
			rss += res[i] * res[i];
		}
//		for (int i = 0; i < p.size(); i += 2) {
//			rss += 1e-4 * p[i] * p[i];
//			rss += 1e-6 * p[i+1] * p[i+1];
//		}
	}
};


LSQGeneticFit::LSQGeneticFit(LSQContext &lsq, int nparams)
{
	this->lsqCont = lsq;
	this->nparams = nparams;
	std_alpha = 0.1;
	std_theta = 50;
	eliteSize = nparams;
	popSize = eliteSize * 10;
	maxIter = 5000;
	verbose = false;
}

std::vector <double> LSQGeneticFit::DataFit()
{
	std::vector <double> param0;
	for (int i = 0; i < nparams; i += 2) {
		param0.push_back(1e-3);
		param0.push_back(1);		
	}
	return DataFit(param0);
}

std::vector <double> LSQGeneticFit::DataFit(std::vector <double> param0)
{
	std::vector <ParamVector> pop(popSize);	
	// Initialize random number generator
	srand(time(NULL));
	for (int i = 0; i < popSize; i++) {
		pop[i].SetParam(param0);
		if (i > 0) {
			pop[i].Mutate(std_alpha, std_theta);
		}
		pop[i].CalcRSS(lsqCont);
	}
	// Iterations of genetic algorithm	
	int iter = 0, iter_nochange = 0;
	double rss_old = 1e200, rss = 1e300;
	while (iter < maxIter && iter_nochange < maxIter / 10) {
		iter++;
		rss_old = rss;
		// Sort the population
		std::sort(pop.begin(), pop.end());
		// Estimate some variables for mutations
		double std_alpha_vechuge[nparams];
		double std_theta_vechuge[nparams];
		double std_alpha_veclarge[nparams];
		double std_theta_veclarge[nparams];
		double std_alpha_vecsmall[nparams];
		double std_theta_vecsmall[nparams];
		for (int j = 0; j < nparams; j += 2) {
			std_alpha_vechuge[j/2] = pop[0].p[j] * 2;
			std_theta_vechuge[j/2] = pop[0].p[j+1] * 2;
			std_alpha_veclarge[j/2] = pop[0].p[j] / 3.0;
			std_theta_veclarge[j/2] = pop[0].p[j+1] / 3.0;
			std_alpha_vecsmall[j/2] = pop[0].p[j] / 100.0;
			std_theta_vecsmall[j/2] = pop[0].p[j+1] / 100.0;
		}
		// Create the new generation
		for (int i = eliteSize; i < popSize; i++) {
			int p1ind = rand() % eliteSize;
			int p2ind = rand() % eliteSize;
			pop[i].Cross(pop[p1ind], pop[p2ind]);
			
			double mutProb = (double) rand() / RAND_MAX;

			
			if (mutProb < 0.1) {
//				pop[i].Mutate(std_alpha, std_theta);
				pop[i].Mutate(std_alpha_vechuge, std_theta_vechuge);
			} else if (mutProb < 0.15) {
				pop[i].Mutate(std_alpha_veclarge, std_theta_veclarge);
			} else if (mutProb < 0.25) {
				pop[i].Mutate(std_alpha_vecsmall, std_theta_vecsmall);
			}

			pop[i].CalcRSS(lsqCont);
		}
		rss = pop[0].rss;
		if (rss < rss_old) {
			iter_nochange = 0;
			if (verbose) {
				printf("%d %f | ", iter, pop[0].rss);
				for (int i = 0; i < pop[0].p.size(); i++) {
					printf("%f ", pop[0].p[i]);
				}
				printf("\n");
			}
		} else {
			iter_nochange++;
		}
	}
	return pop[0].p;
}

